from abc import ABC, abstractmethod
from random import randint
from typing import Set, Iterable, Type


class DiceGameBase(ABC):
    def __init__(self, dice_board: "Dice", player_name: str):
        self._dice_board = dice_board
        self._player_name = player_name

    @abstractmethod
    def check_winner(self) -> bool:
        ...

    @abstractmethod
    def roll(self) -> None:
        ...

    @abstractmethod
    def add_dice(self, die: Type["DieBase"]) -> None:
        ...

    @property
    @abstractmethod
    def total_points(self) -> int:
        ...


class DieBase(ABC):
    def __init__(self) -> None:
        self.face = None
        self.roll()

    @abstractmethod
    def roll(self) -> None:
        ...

    def __repr__(self) -> str:
        return str(self.face)


class D4(DieBase):
    def roll(self) -> None:
        self.face = randint(1, 4)


class D6(DieBase):
    def roll(self) -> None:
        self.face = randint(1, 6)


class D8(DieBase):
    def roll(self) -> None:
        self.face = randint(1, 8)


class Dice:

    def __init__(self, *die_classes) -> None:
        self.__dice = [dc() for dc in die_classes]
        self.__adjust: int = 0
        self.rolled_pos: Set[int] = set()

    @property
    def total(self) -> int:
        return sum(d.face for d in self.__dice) + self.__adjust

    @property
    def dice(self):
        return self.__dice

    @dice.setter
    def dice(self, value):
        raise NotImplementedError

    def plus(self, adjust: int = 0):
        self.__adjust = adjust
        return self

    def roll(self) -> None:
        if not self.rolled_pos:
            for die in self.dice:
                die.roll()
        else:
            for i in self.rolled_pos:
                self.__dice[i].roll()
        self.rolled_pos = set()

    def to_be_rolled(self, pos: Iterable[int]):
        if not all(0 <= n < len(self.__dice) for n in pos):
            raise IndexError("Some indices are out of range")
        self.rolled_pos = set(pos)
        return self

    def __repr__(self):
        return str(self.__dice)

    def __add__(self, other):
        if issubclass(other, DieBase):
            new_classes = [type(d) for d in self.__dice] + [other]  # D4, D6, D8
            new = Dice(*new_classes)
            return new
        raise ValueError(f"Type {type(other)} not supported")

    def __radd__(self, other):
        return self.__add__(other)

    def __iadd__(self, other):
        if issubclass(other, DieBase):
            self.__dice.append(other())
            return self

    def __mul__(self, other):
        if isinstance(other, int):
            new_classes = [type(d) for d in self.__dice] * other
            new = Dice(*new_classes)
            return new
        raise NotImplementedError(f"Value of type {type(other)} is not supported")

    def __imul__(self, other):
        if isinstance(other, int):
            new_dice = [type(d) for d in self.__dice] * other
            self.__dice = [die() for die in new_dice]
            return self
        raise NotImplementedError(f"Value of type {type(other)} is not supported")

    def __rmul__(self, other):
        raise NotImplementedError



if __package__ is None or __package__ == '':
    # uses current directory visibility
    from game.render_game import RanderGame
else:
    # uses current package visibility
    from .game.render_game import RanderGame


def start():
    game = RanderGame()
    game.play()


if __name__ == "__main__":
    start()



import sys

from .config import color_config as color
from .config import text_constant as text
from .config import game_config as cfg
from .library.abstract_dice import Dice, D4, D6, D8
from .baboo_game import BabooGame


class RanderGame:
    def __init__(self):
        print("\x1b[8;43;98t")  # This line is for resizing user terminal
        print(self.__show_text(text.WELCOME_MESSAGE, color.YELLOW))
        print(self.__show_text(text.DICE_IMAGE, color.BLUE))
        self.__player_name = self.__get_player_name()
        self.__board = Dice(D4, D4)
        self.__game = BabooGame(self.__board, self.__player_name)
        self.__round = 1

    def __get_player_name(self):
        # name = input(self.__show_text("Please insert your name: "))
        name = "STRANGER"
        return name

    def __show_text(self, text, col=color.WHITE):
        return f"{col}{text}{color.RESET}"

    def __continue(self):
        self.__game.roll()

    def __roll_specific(self):
        while True:
            dice = self.__game.dice
            try:
                pos = input(self.__show_text(f"Select positions that you want to keep start from 1 to {len(dice)}: "))
                pos = list(map(lambda x: int(x) - 1, pos.split()))
                if max(pos) > (len(dice) - 1):
                    raise Exception
                else:
                    self.__game.roll(pos)
                    return
            except Exception:
                print(self.__show_text("Please insert a valid position ", color.RED))

    def __add_die(self, die):
        while True:
            position = input(
                self.__show_text("Enter position to replace your die or press enter to append to the end: "))
            try:
                if position == "":
                    print("empty")
                    self.__game.add_dice(die)
                    return
                else:
                    self.__game.dice[int(position) - 1] = die()
                    return
            except Exception:
                print(self.__show_text(f"Wrong position please please select between 1 and {len(self.__game.dice)}",
                                       color.RED))

    def __show_die_menu(self):
        print(self.__show_text(f"""
1. D4({cfg.DIE4_COST} Cr.)     2. D6({cfg.DIE6_COST} Cr.)      3. D8({cfg.DIE8_COST} Cr.)      4. Exit
        """))
        current_balance = self.__game.credit_balance
        while True:
            menu_number = input(self.__show_text("Select Die you want to buy: "))
            try:
                if 1 <= int(menu_number) <= 4:
                    match int(menu_number):
                        case 1:
                            if current_balance >= cfg.DIE4_COST + cfg.ROLL_COST:
                                self.__game.credit_balance -= cfg.DIE4_COST
                                self.__add_die(D4)
                                break
                        case 2:
                            if current_balance >= cfg.DIE6_COST + cfg.ROLL_COST:
                                self.__game.credit_balance -= cfg.DIE6_COST
                                self.__add_die(D6)
                                break
                        case 3:
                            if current_balance >= cfg.DIE8_COST + cfg.ROLL_COST:
                                self.__game.credit_balance -= cfg.DIE8_COST
                                self.__add_die(D8)
                                break
                        case 4:
                            break
                else:
                    raise Exception
            except Exception:
                print(self.__show_text("Select a valid menu number between 1 and 4: ", color.RED))

    def __withdraw(self):
        current_balance = self.__game.credit_balance
        print(self.__show_text(text.CONGRATULATIONS % current_balance, color.GREEN))
        # print(self.__show_text(WIN_AMOUNT % current_balance, color.GREEN))
        sys.exit()

    def __show_menu(self):

        print(self.__show_text("""
1. Continue      2. Roll specific dice      3. Add new die     4. Withdraw
        """))

        while True:
            menu_number = input(self.__show_text("Please chose a menu number: "))
            try:
                if 1 <= int(menu_number) <= 4:
                    match int(menu_number):
                        case 1:
                            self.__continue()
                        case 2:
                            self.__roll_specific()
                        case 3:
                            self.__show_die_menu()
                        case 4:
                            self.__withdraw()
                    break
                else:
                    raise Exception

            except Exception:
                print(self.__show_text("Please select a menu number between 1-4", color.RED))

    def __display_board(self):
        dice_list = self.__game.dice

        print(self.__show_text(f"\n{'▔' * 76}"))
        print(self.__show_text("Round:"), self.__show_text(self.__round, color.GREEN), end="    ")

        status_lose = self.__show_text("Lose", color.RED)
        status_win = self.__show_text("Win", color.GREEN)

        if self.__game.check_winner():
            print(self.__show_text("Status:"), status_win, end="     ")
        else:
            print(self.__show_text("Status:"), status_lose, end="     ")

        print(self.__show_text(f"Balance: {self.__game.credit_balance}"), end="     ")
        print(end="\n\n")

        for die in dice_list:
            print("    ", self.__show_text(die.face, color.Fore.LIGHTCYAN_EX), end="       ")
        print()

        for line in range(5):
            for die in dice_list:
                print(self.__show_text(text.DICE.get(die.face)[line], color.BLUE), end="  ")
            print()

        for die in dice_list:
            print("    ", self.__show_text(type(die).__name__, color.MAGENTA), end="      ")
        print(end="\n")

        print(self.__show_text(f"{'▁' * 76}"), end="\n\n")

    def play(self):
        input("Press ENTER to start the game: ")
        print(self.__show_text(text.START_GAME))
        self.__game.credit_balance -= cfg.ROLL_COST
        while self.__game.credit_balance >= cfg.ROLL_COST:
            if self.__game.check_winner():
                self.__game.credit_balance += self.__game.total_points
                self.__display_board()
                self.__game.roll()
                self.__show_menu()
                self.__round += 1
            else:
                self.__display_board()
                self.__show_menu()
                self.__round += 1
        print(self.__show_text(text.GAME_OVER, color.RED))



from .library.abstract_dice import *
from .config.game_config import *


class BabooGame(DiceGameBase):
    def __init__(self, dice_board: "Dice", player_name: str):
        super().__init__(dice_board, player_name)
        self._credit_balance = INITIAL_BALANCE

    @property
    def dice(self):
        return self._dice_board.dice

    @property
    def player_name(self):
        return self._player_name

    @property
    def total_points(self) -> int:
        return self._dice_board.plus(len(self._dice_board.dice)).total * ROLL_COST

    @property
    def credit_balance(self):
        return self._credit_balance

    @credit_balance.setter
    def credit_balance(self, amount):
        self._credit_balance = amount

    def check_winner(self) -> bool:
        return len(set(die.face for die in self._dice_board.dice)) == 1

    def roll(self, roll_pos: Iterable[int] = None) -> None:
        if self._credit_balance >= ROLL_COST:
            if roll_pos:
                self._dice_board.to_be_rolled(roll_pos).roll()
            else:
                self._dice_board.roll()

            self._credit_balance -= ROLL_COST
        else:
            raise PermissionError("Not enough balance")

    def add_dice(self, die: Type["DieBase"]):
        self._dice_board += die

    def replace_die(self, pos, die):
        self._dice_board[pos] = die()



from colorama import Fore

RED = Fore.LIGHTRED_EX
GREEN = Fore.LIGHTGREEN_EX
RESET = Fore.RESET
YELLOW = Fore.LIGHTYELLOW_EX
WHITE = Fore.LIGHTWHITE_EX
BLACK = Fore.LIGHTBLACK_EX
BLUE = Fore.LIGHTBLUE_EX
MAGENTA = Fore.LIGHTMAGENTA_EX
